'use strict';

var jf = require('jsonfile');
var fse = require('fs-extra');
var klaw = require('klaw');
var path = require('path');
var through2 = require('through2');

var _ = require('lodash');
var asyncLib = require('async');

var moduleConfig = {
    CONFIG_NAME: 'interface-enum',
    MARK_NAME: 'generateInterfaceToEnum',
    ENUM_FILE_SUFFIX: '.interfaceEnums.ts',
    FILE_EXTENSION: '.ts',
    MARK_INTERFACES: '.*generateInterfaceToEnum[\\s\\S]*?^^export interface (\\w+)'
};

var TokenType = {
    END: '}',
    START: '{',
    ASSIGN: ':',
    NON_MANDATORY: '?',
    END_OF_LINE: ';'
};

/**
 * Define interface object
 * @typedef {Object} InterfaceObject
 * @property {String} name
 * @property {Object} interfaceObject
 */

/**
 * Define file interface object
 * @typedef {Object} FileInterfaceObject
 * @property {String} path
 * @property {InterfaceObject[]} interfacesObjects
 */
// generateInterfaceToEnum

// we should only iterate over typescript files
var excludeNotTs = through2.obj(function excludeNotTsFunction(item, enc, next) {
    if (path.extname(item.path) === moduleConfig.FILE_EXTENSION) this.push(item);

    next();
});

var markFileRegEx = new RegExp(moduleConfig.MARK_NAME, 'g');
var markInterfacesRegEx = new RegExp(moduleConfig.MARK_INTERFACES, 'gm');

var filesFound = [];
var filesToProcess = [];
var filesToCreate = [];
var configDataObject = void 0;

module.exports = function tsInterfaceEnum(configOptionsData, cb) {
    var jobs = [loadConfig, iterateConfigPaths, checkFiles, removeOldEnumFiles, processFiles, createFiles];

    if (!cb) {
        cb = configOptionsData;
        configOptionsData = null;
    }

    if (configOptionsData) configDataObject = configOptionsData;

    asyncLib.waterfall(jobs, function (err) {
        if (err) {
            return cb(err);
        }

        cb();
        console.log('All enum file generated.');
    });
};

function loadConfig(next) {
    if (configDataObject) {
        return next();
    }

    jf.readFile('./' + moduleConfig.CONFIG_NAME + '.json', function (err, data) {
        if (err) return next('Error loading config file.');

        configDataObject = data;
        next();
    });
}

function iterateConfigPaths(next) {
    if (_.isArray(configDataObject.paths)) {
        asyncLib.eachSeries(configDataObject.paths, iteratePath, function (err) {
            next(err);
        });
    }
}

/**
 * walk over given path
 * only select .ts files
 * @param {String} singlePathString
 * @param {Function} nextPath
 */
function iteratePath(singlePathString, nextPath) {
    klaw(singlePathString).pipe(excludeNotTs).on('data', function (item) {
        filesFound.push(item.path);
    }).on('error', function (err, item) {
        console.log('Error loading file: ', item.path); // the file the error occurred on
    }).on('end', function () {
        nextPath();
    });
}

function checkFiles(next) {
    asyncLib.eachSeries(filesFound, markFile, function (err) {
        next(err);
    });
}
/**
 * check file for presence of lib interface prefix
 * @param {String} singlePathString
 * @param {Function} nextPath
 */
function markFile(singlePathString, nextPath) {
    // reset global index
    markFileRegEx.lastIndex = 0;
    fse.readFile(singlePathString, function (err, fileBuffer) {
        if (err) return nextPath(err);

        if (markFileRegEx.test(fileBuffer.toString())) {
            filesToProcess.push(singlePathString);
        }

        nextPath();
    });
}

function removeOldEnumFiles(next) {
    asyncLib.eachSeries(filesToProcess, removeOldEnumFile, function (err) {
        next(err);
    });
}

/**
 * remove all file we are creating this run
 * @param {String} singlePathString
 * @param {Function} nextFile
 */
function removeOldEnumFile(singlePathString, nextFile) {
    var enumFileData = getEnumFileData(singlePathString);

    fse.remove(enumFileData.enumFileFull, function (err) {
        nextFile(err);
    });
}

function processFiles(next) {
    asyncLib.eachSeries(filesToProcess, processFile, function (err) {
        next(err);
    });
}

/**
 * generate interface object array per file
 * @param {String} singlePathString
 * @param {Function} nextPath
 */
function processFile(singlePathString, nextPath) {
    var interfacesObjects = [];

    var processJobs = [_findInterfaces, _generateObjects];

    asyncLib.waterfall(processJobs, function (err) {
        var ans = { path: singlePathString, interfacesObjects: interfacesObjects };

        filesToCreate.push(ans);

        nextPath(err, ans);
    });

    /**
     * generate interface
     * @param nextTask
     */
    function _findInterfaces(nextTask) {
        findInterfaces(singlePathString, function (err, interfaces) {
            if (err) return nextTask(err);

            nextTask(null, interfaces);
        });
    }

    function _generateObjects(interfaces, nextTask) {
        asyncLib.eachSeries(interfaces, _generateObject, function (err) {
            nextTask(err);
        });
    }

    function _generateObject(singleInterface, nextInterface) {
        generateObject(singlePathString, singleInterface, function (err, interfaceObject) {
            if (err) return nextInterface(err);

            if (interfaceObject) interfacesObjects.push(interfaceObject);

            nextInterface();
        });
    }
}
/**
 * return array of interfaces found
 * @param {String} singlePathString
 * @param {Function} cb - err or string[]
 */
function findInterfaces(singlePathString, cb) {
    var interfaces = [];
    fse.readFile(singlePathString, function (err, fileBuffer) {
        if (err) return cb(err);

        var fileString = fileBuffer.toString();
        var foundMatch = markInterfacesRegEx.exec(fileString);

        while (foundMatch) {
            if (foundMatch.length && foundMatch.length >= 2) {
                interfaces.push(foundMatch[1]);
            }

            foundMatch = markInterfacesRegEx.exec(fileString);
        }

        cb(null, interfaces);
    });
}
/**
 * Find the interface start in file
 * And send it to farther processing
 * @param {String} singlePathString
 * @param {String} singleInterface
 * @param {Function}nextInterface
 */
function generateObject(singlePathString, singleInterface, nextInterface) {
    fse.readFile(singlePathString, function (err, fileBuffer) {
        if (err) return nextInterface(err);

        var fileString = fileBuffer.toString();

        var lines = fileString.split('\n');

        var interfaceReg = new RegExp('^export interface ' + singleInterface.trim());
        var interfaceLine = null;
        var interfaceFound = false;

        lines.forEach(function (singleLine, index) {
            if (singleLine.match(interfaceReg)) {
                interfaceFound = true;
                interfaceLine = index;
            }
        });

        if (interfaceFound) {
            var slicedArray = _.takeRight(lines, lines.length - interfaceLine);
            var tmpString = slicedArray.join('\n');

            var interfaceObject = parseInterfaceToObject(tmpString);

            if (_.isEmpty(interfaceObject)) {
                nextInterface();
            } else {
                nextInterface(null, { name: singleInterface, interfaceObject: interfaceObject });
            }
        } else {
            nextInterface();
        }
    });
}

/**
 * generate js object from found interface
 * only take first level
 * only take first object
 * @param {String} interfaceCandidate
 * @return {Object} found interface object
 */
function parseInterfaceToObject(interfaceCandidate) {
    var objectsFound = []; // start, end, include
    // iterateLines
    var stringLines = interfaceCandidate.split('\n');

    for (var i = 0; i < stringLines.length; i++) {
        for (var j = 0; j < stringLines[i].length; j++) {
            if (stringLines[i][j] === TokenType.START) {
                objectsFound.push({ start: { col: j, line: i } });
            }

            if (stringLines[i][j] === TokenType.END) {
                if (objectsFound.length) {
                    for (var k = 1; k <= objectsFound.length; k++) {
                        var foundObject = objectsFound[objectsFound.length - k];

                        if (foundObject && !foundObject.end) {
                            foundObject.end = { col: j, line: i };
                            break;
                        }
                    }
                }
            }
        }

        // top level interface is inside
        if (objectsFound.length) {
            if (objectsFound[0].hasOwnProperty('start') && objectsFound[0].hasOwnProperty('end')) {
                break;
            }
        }
    }

    var objectInterface = {};

    if (objectsFound.length) {
        if (objectsFound[0].hasOwnProperty('start') && objectsFound[0].hasOwnProperty('end')) {
            var _loop = function _loop(_i) {
                var found = _.findLastIndex(objectsFound, function (singleElement) {
                    return _i > singleElement.start.line && _i < singleElement.end.line;
                });

                if (found === 0 || !found) {
                    var trimLine = stringLines[_i].trim();
                    var keyName = trimLine.match(/^(\w+)/g);

                    if (keyName) {
                        objectInterface[keyName[0]] = keyName[0];
                    }
                }
            };

            // as for now only top level are in our interest
            for (var _i = objectsFound[0].start.line + 1; _i <= objectsFound[0].end.line - 1; _i++) {
                _loop(_i);
            }
        }
    }

    return objectInterface;
}

function createFiles(next) {
    asyncLib.eachSeries(filesToCreate, createFile, function (err) {
        next(err);
    });
}
/**
 * Create enums and file
 * @param {FileInterfaceObject} fileToCreateObject
 * @param nextFile
 */
function createFile(fileToCreateObject, nextFile) {
    var enumList = fileToCreateObject.interfacesObjects.map(function (singleInterfaceObject) {
        var stringToReturn = convertObjectToEnumString(singleInterfaceObject.name, singleInterfaceObject.interfaceObject);

        return '' + stringToReturn;
    });

    var enumsString = enumList.join('\n\n');

    var enumFileData = getEnumFileData(fileToCreateObject.path);

    var fileString = '// This file was generated by TsInterfaceEnum\n// ' + enumFileData.enumFileName + '\n\n' + enumsString + '\n';
    fse.outputFile(enumFileData.enumFileFull, fileString, function (err) {
        nextFile(err);
    });
}
/**
 *
 * @param {String} name
 * @param {Object}  objectToCreate
 * @return {string}
 */
function convertObjectToEnumString(name, objectToCreate) {
    var keysStringKeys = Object.keys(objectToCreate);
    var keysStringArray = keysStringKeys.map(function (key, index) {
        if (index + 1 === keysStringKeys.length) {
            return '\t' + key + ' = \'' + objectToCreate[key] + '\' as any';
        }
        return '\t' + key + ' = \'' + objectToCreate[key] + '\' as any,';
    });

    var allKeyString = keysStringArray.join('\n');

    return '\n// this enum was auto generated\nexport enum ' + name + 'Enum {\n' + allKeyString + '\n}';
}

/**
 * generate enum files name
 * @param {String} filePath
 * @return {{enumFileFull: string, enumFileName: string}}
 */
function getEnumFileData(filePath) {
    var fileName = path.basename(filePath, moduleConfig.FILE_EXTENSION);
    var dirName = path.dirname(filePath);
    var enumFileName = fileName + moduleConfig.ENUM_FILE_SUFFIX;

    return {
        enumFileFull: dirName + '/' + enumFileName,
        enumFileName: enumFileName
    };
}